"""
The module creates the inflation scenario associated to a causal structure. See
arXiv:1609.00672 and arXiv:1707.06476 for the original description of
inflation.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import warnings
from collections import defaultdict
from functools import reduce, cached_property
from itertools import (chain,
                       combinations_with_replacement,
                       permutations)
from typing import Tuple, List, Union, Dict
from warnings import warn

import networkx as nx
import numpy as np
from networkx.algorithms import isomorphism
from sympy import Symbol
from tqdm import tqdm
from .sdp.fast_npa import (nb_classify_disconnected_components,
                           nb_overlap_matrix,
                           apply_source_perm,
                           commutation_matrix)
from .symmetry_utils import group_elements_from_generators
from .utils import (format_permutations,
                    partsextractor,
                    perm_combiner,
                    all_and_maximal_cliques)

# Force warnings.warn() to omit the source code line in the message
# https://stackoverflow.com/questions/2187269/print-only-the-message-on-warnings
formatwarning_orig = warnings.formatwarning
warnings.formatwarning = lambda msg, category, filename, lineno, line=None: \
    formatwarning_orig(msg, category, filename, lineno, line="")


class InflationProblem:
    """Class for encoding relevant details concerning the causal compatibility.
    """
    def __init__(self,
                 dag: Union[Dict, None]=None,
                 outcomes_per_party: Union[Tuple[int,...], List[int], np.ndarray]=tuple(),
                 settings_per_party: Union[Tuple[int,...], List[int], np.ndarray]=tuple(),
                 inflation_level_per_source: Union[Tuple[int,...], List[int], np.ndarray]=tuple(),
                 classical_sources: Union[str, Tuple[str,...], List[str], None]=tuple(),
                 nonclassical_intermediate_latents: Union[Tuple[str,...], List[str]]=tuple(),
                 classical_intermediate_latents: Union[Tuple[str,...], List[str]]=tuple(),
                 order: Union[Tuple[str,...], List[str]]=tuple(),
                 verbose=0):
        """Class for encoding relevant details concerning the causal compatibility
        scenario.

        Parameters
        ----------
        dag : Dict[str, List[str]], optional
            Dictionary where each key is a parent node, and the corresponding value
            is a list of the corresponding children nodes. By default it is a
            single source connecting all the parties.
        outcomes_per_party : [np.ndarray, List[int], Tuple[int,...]]
            Measurement outcome cardinalities.
        settings_per_party : [np.ndarray, List[int], Tuple[int,...]], optional
            Measurement setting cardinalities. By default ``1`` for all parties.
        inflation_level_per_source : [int, List[int]], optional
            Number of copies per source in the inflated graph. Source order is the
            same as insertion order in `dag`. If an integer is provided, it is used
            as the inflation level for all sources. By default ``1`` for all
            sources.
        classical_sources : Union[List[str], str], optional
            Names of the sources that are assumed to be classical. If ``'all'``,
            it imposes that all sources are classical. By default an empty tuple.
        nonclassical_intermediate_latents: Tuple[str], optional
            Designates non-exogynous nodes in the DAG with these names as
            nonclassical latent, in addition to all exogynous nodes other than
            those specifically indicates as classical.
        classical_intermediate_latents: Tuple[str], optional
            Designates non-exogynous nodes in the DAG with these names as
            a classical-type latent, alongside any latents from classical_sources.
        order : Tuple[str], optional
            Name of each party. This also fixes the order in which party outcomes
            and settings are to appear in a conditional probability distribution.
            Default is alphabetical order and labels, e.g., ``['A', 'B', ...]``.
        verbose : int, optional
            Optional parameter for level of verbose:

            * 0: quiet (default),
            * 1: monitor level: track program process and show warnings,
            * 2: debug level: show properties of objects created.
        """
        self.verbose = verbose

        # Read settings and outcomes
        if not outcomes_per_party:
            raise ValueError("Please provide outcomes per party.")
        self.outcomes_per_party = np.array(outcomes_per_party, dtype=int)
        self.nr_parties         = len(self.outcomes_per_party)
        if not settings_per_party:
            if self.verbose > 0:
                warn("No settings per party provided, " +
                     "assuming all parties have one setting.")
            self.private_settings_per_party = np.ones(self.nr_parties,
                                                      dtype=int)
        else:
            self.private_settings_per_party = np.asarray(settings_per_party,
                                                         dtype=int)
            assert len(self.private_settings_per_party) == self.nr_parties, \
                (f"You have specified a list of {len(outcomes_per_party)} "
                 + f"outcomes and a list of {len(settings_per_party)} inputs. "
                 + "These lists must have the same length and equal to the "
                 + "number of visible variables in the scenario.")

        # Record expected shape of input distributions
        self.expected_distro_shape = tuple(np.hstack(
            (self.outcomes_per_party,
             self.private_settings_per_party)).tolist())

        # Internal record of intermediate latents
        self.classical_intermediate_latents = set(map(str,classical_intermediate_latents))
        self.nonclassical_intermediate_latents = set(map(str,nonclassical_intermediate_latents))
        assert self.classical_intermediate_latents.isdisjoint(self.nonclassical_intermediate_latents), "An intermediate latent cannot be both classical and nonclassical."
        self.intermediate_latents = self.classical_intermediate_latents.union(self.nonclassical_intermediate_latents)

        # Assign names to the visible variables
        names_have_been_set_yet = False
        if dag:
            implicit_names = set(map(str, chain.from_iterable(dag.values()))).difference(self.intermediate_latents)
            assert len(implicit_names) == self.nr_parties, \
                ("You must provide a number of outcomes for the following "
                 + f"{len(implicit_names)} variables: {implicit_names}")
            if order:
                sanity_check = (implicit_names.issubset(order)
                                and implicit_names.issuperset(order))
                if sanity_check:
                    self.names = order
                    names_have_been_set_yet = True
                else:
                    if self.verbose > 0:
                        warn("The names read from the DAG do not match those "
                             + "read from the `order` argument. The names used"
                             + " are those read from the DAG.")
            if not names_have_been_set_yet:
                if len(implicit_names) > 1:
                    if self.verbose > 0:
                        warn("The order of variables is inferred by the DAG "
                             + "according to lexicographic order.")
                self.names = sorted(implicit_names)
                names_have_been_set_yet = True
        if order and (not names_have_been_set_yet):
            sanity_check = (len(order) == self.nr_parties)
            if sanity_check:
                self.names = tuple(map(str, order))
                names_have_been_set_yet = True
            else:
                if self.verbose > 0:
                    warn("The number of names provided does not match the "
                         + "number of variables that need a name. The names "
                         + "used are determined by the list of variables "
                         + "with outcomes.")
        if not names_have_been_set_yet:
            self.names = [chr(ord("A") + i) for i in range(self.nr_parties)]

        if not dag:
            if self.verbose > 0:
                warn("The DAG must be a non-empty dictionary with parent "
                     + "variables as keys and lists of children as values. "
                     + "Defaulting to one global source.")
            self.dag = {"h_global": set(self.names)}
        else:
            # Sources are nodes with children but not parents
            self.dag = {str(parent): set(map(str, children)) for
                        parent, children in dag.items()}


        # Distinguishing between classical versus nonclassical sources
        nodes_with_children_as_list = list(self.dag.keys())
        nodes_with_children = set(nodes_with_children_as_list)
        self._actual_sources = np.asarray(sorted(
            nodes_with_children.difference(self.names, self.intermediate_latents),
            key=nodes_with_children_as_list.index), dtype=object)
        self.nr_sources = len(self._actual_sources)
        if isinstance(classical_sources, str):
            if classical_sources.lower() == "all":
                self._classical_sources = np.ones(self.nr_sources, dtype=bool)
            else:
                raise ValueError(f'The keyword argument classical_sources=`{classical_sources}` could not be parsed.')
        else:
            self._classical_sources = np.zeros(self.nr_sources, dtype=bool)
        if not isinstance(classical_sources, (str, type(None))):
            assert set(classical_sources).issubset(self._actual_sources), "Some specified classical source cannot be found in the DAG."
            for ii, source in enumerate(self._actual_sources):
                if source in classical_sources:
                    self._classical_sources[ii] = 1
        self._nonclassical_sources = np.logical_not(self._classical_sources)

        self._inverse_dag = defaultdict(set)
        for v, children in self.dag.items():
            for child in children:
                self._inverse_dag[child].add(v)
        for il in self.intermediate_latents:
            this_ils_parents = self._inverse_dag[il]
            if not this_ils_parents.isdisjoint(self.names):
                raise NotImplementedError(
                    "InflationProblem cannot handle intermediate latents with observable parents at this time.")
            elif this_ils_parents.isdisjoint(self._actual_sources.flat):
                raise ValueError(
                    f"The intermediate latent {il} has no source parent. Please add a source parent to the DAG and re-initialize InflationProblem.")

        for ncil in self.nonclassical_intermediate_latents:
            this_ils_parents = self._inverse_dag[ncil]
            if this_ils_parents.isdisjoint(self._actual_sources[self._nonclassical_sources].flat):
                raise NotImplementedError(
                    f"The nonclassical intermediate latent {ncil} has no nonclassical source parent." +
                    f"\nIts parents are {this_ils_parents} while the set of nonclassical sources is {self._actual_sources[self._nonclassical_sources]})" +
                    "\nPlease add a nonclassical source parent to the DAG and re-initialize InflationProblem.")



        # Unpacking of visible nodes with children
        parties_with_children = nodes_with_children.intersection(self.names)
        self.has_children   = np.zeros(self.nr_parties, dtype=bool)
        self.is_network     = set(nodes_with_children).isdisjoint(self.names)
        names_to_integers = {party: position
                             for position, party in enumerate(self.names)}
        adjacency_matrix = np.zeros((self.nr_parties, self.nr_parties),
                                    dtype=bool)
        for parent in parties_with_children:
            ii = names_to_integers[parent]
            self.has_children[ii] = True
            observable_children = set(self.dag[parent])
            if not observable_children.issubset(self.names):
                raise NotImplementedError("At this time InflationProblem does not accept DAGs with observed nodes pointing to intermediate latents.")
            latents_behind_this_node = self._inverse_dag[parent].difference(self.names)
            siblings_of_parent = set([])
            for latent in latents_behind_this_node:
                siblings_of_parent.update(self.dag[latent])
            if not observable_children.issubset(siblings_of_parent):
                raise NotImplementedError(
                    "At this time InflationProblem does not accept DAGs with directed edges between observed nodes lacking a common latent parent.")
            child_indices = [names_to_integers[child] for child in observable_children]
            adjacency_matrix[ii, child_indices] = True

        # Compute number of settings for the unpacked variables
        self.parents_per_party = list(map(np.flatnonzero, adjacency_matrix.T))
        settings_per_party_lst = [[s] for s in self.private_settings_per_party]
        for party_idx, party_parents_idxs in enumerate(self.parents_per_party):
            settings_per_party_lst[party_idx].extend(
                np.take(self.outcomes_per_party, party_parents_idxs))
        self.settings_per_party = np.asarray(
            [np.prod(multisetting) for multisetting in settings_per_party_lst],
            dtype=int)

        # Build the correspondence between effective settings and the true
        # tuple of setting values of all parents.
        effective_to_parent_settings = []
        for i in range(self.nr_parties):
            effective_to_parent_settings.append(dict(zip(
                range(self.settings_per_party[i]),
                np.ndindex(*settings_per_party_lst[i]))))
        self.effective_to_parent_settings = effective_to_parent_settings

        # Create network corresponding to the unpacked scenario
        self.hypergraph = np.zeros((self.nr_sources, self.nr_parties),
                                   dtype=np.uint8)
        # We need to distinguish between quantum versus classical sources.
        # We use the internal convention that +1 indicates a shared classical source
        # whereas +2 indicates a shared quantum sources
        self.sources_to_check_for_party_pair_commutation = np.zeros((self.nr_parties, self.nr_parties, self.nr_sources),
                                    dtype=np.uint8)
        for source_idx, source in enumerate(self._actual_sources):
            quantum_source_bonus = self._nonclassical_sources[source_idx]
            children = self.dag[source]
            observable_children = children.intersection(self.names)
            latent_children = children.difference(self.names)
            assert latent_children.issubset(self.intermediate_latents), f"{latent_children.difference(self.intermediate_latents)} are not all a recognized party or an intermediate latent."
            for child in observable_children:
                child_idx = names_to_integers[child]
                self.hypergraph[source_idx, child_idx] = 1
                self.sources_to_check_for_party_pair_commutation[child_idx, child_idx, source_idx] = 1 + quantum_source_bonus
            for intermediate_latent in latent_children:
                observable_descendants_via_this_latent = self.dag[intermediate_latent]
                assert observable_descendants_via_this_latent.issubset(self.names), "At this time InflationProblem does not accept DAGs with intermediate latents pointing to other intermediate latents."
                quantum_connection_bonus = np.logical_and(intermediate_latent in self.nonclassical_intermediate_latents, quantum_source_bonus).astype(np.uint8)
                for desc in observable_descendants_via_this_latent:
                    desc_idx = names_to_integers[desc]
                    self.hypergraph[source_idx, desc_idx] = 1
                    for desc2 in observable_descendants_via_this_latent:
                        desc2_idx = names_to_integers[desc2]
                        self.sources_to_check_for_party_pair_commutation[desc_idx, desc2_idx, source_idx] = max(
                        self.sources_to_check_for_party_pair_commutation[desc_idx, desc2_idx, source_idx],
                            1 + quantum_connection_bonus)


        assert np.sum(self.hypergraph, axis=0).all(), \
            ("There appears to be a party with no sources in its past. This is not allowed.")

        if not inflation_level_per_source:
            if self.verbose > 0:
                warn("The inflation level per source must be a non-empty list."
                     + " Defaulting to 1 (no inflation, just NPA hierarchy).")
            self.inflation_level_per_source = np.array([1] * self.nr_sources)
        elif type(inflation_level_per_source) == int:
            self.inflation_level_per_source = \
                np.array([inflation_level_per_source] * self.nr_sources)
        else:
            self.inflation_level_per_source = \
                np.array(inflation_level_per_source)
            assert self.nr_sources == len(self.inflation_level_per_source), \
                ("The number of sources as described by the unpacked " +
                 f"hypergraph, {self.nr_sources}, and by the list of " +
                 "inflation levels specified, " +
                 f"{len(self.inflation_level_per_source)}, does not coincide.")

        # Determine if the inflation problem has a factorizing pair of parties.
        shared_sources = [np.all(np.vstack(pair), axis=0) for pair in
                          combinations_with_replacement(self.hypergraph.T, 2)]
        just_one_copy = (self.inflation_level_per_source == 1)
        self.ever_factorizes = False
        for sources_are_shared in shared_sources:
            # If for some two parties, the sources that they share in common
            # can all have different inflation levels, then there exists the
            # potential for factorization.
            if ((not np.any(sources_are_shared))
                or (not np.all(just_one_copy[sources_are_shared]))):
                self.ever_factorizes = True
                break

        # Establish internal dtype
        self._np_dtype = np.result_type(*[
            np.min_scalar_type(np.max(self.settings_per_party)),
            np.min_scalar_type(np.max(self.outcomes_per_party)),
            np.min_scalar_type(self.nr_parties + 1),
            np.min_scalar_type(np.max(self.inflation_level_per_source) + 1)])

        # Create all the different possibilities for inflation indices
        self.inflation_indices_per_party = list()
        for party in range(self.nr_parties):
            inflation_indices = list()
            active_sources =self.hypergraph[:, party]
            num_copies = np.multiply(active_sources,
                                     self.inflation_level_per_source)
            # Put non-participating and non-inflated on equal footing
            num_copies = np.maximum(num_copies, 1)
            for increase_from_base in np.ndindex(*num_copies):
                inflation_indxs = active_sources + np.array(
                    increase_from_base, dtype=self._np_dtype)
                inflation_indices.append(inflation_indxs)
            self.inflation_indices_per_party.append(
                np.vstack(inflation_indices))

        self._all_unique_inflation_indices = np.unique(
            np.vstack(self.inflation_indices_per_party),
            axis=0).astype(self._np_dtype)
        
        # Create hashes and overlap matrix for quick reference
        self._inflation_indices_hash = {op.tobytes(): i for i, op
                                        in enumerate(
                self._all_unique_inflation_indices)}
        self._inflation_indices_overlap = nb_overlap_matrix(
            np.asarray(self._all_unique_inflation_indices, dtype=self._np_dtype))

        # Create the measurements (formerly generate_parties)
        self._nr_properties = 1 + self.nr_sources + 2
        self._astuples_dtype = np.dtype(list(zip(['']*self._nr_properties, [self._np_dtype]*self._nr_properties)))
        self.measurements = list()
        for p in range(self.nr_parties):
            O_vals = np.arange(self.outcomes_per_party[p],
                               dtype=self._np_dtype)
            S_vals = np.arange(self.settings_per_party[p],
                               dtype=self._np_dtype)
            I_vals = self.inflation_indices_per_party[p]
            # TODO: Use broadcasting instead of nested for loops
            measurements_per_party = np.empty(
                (len(I_vals), len(S_vals), len(O_vals), self._nr_properties),
                dtype=self._np_dtype)
            measurements_per_party[:, :, :, 0] = p + 1
            for i, inf_idxs in enumerate(I_vals):
                measurements_per_party[i, :, :, 1:(self.nr_sources + 1)] = \
                    inf_idxs
                for s in S_vals.flat:
                    measurements_per_party[i, s, :, -2] = s
                    for o in O_vals.flat:
                        measurements_per_party[i, s, o, -1] = o
            self.measurements.append(measurements_per_party)

        self.measurements_symbolic = \
            [np.apply_along_axis(self._1d_to_symbol, -1, measurements_per_party)
             for measurements_per_party in self.measurements]

        # Useful for LP
        self._ortho_groups_per_party = []
        for p, measurements_per_party in enumerate(self.measurements):
            O_card = self.outcomes_per_party[p]
            self._ortho_groups_per_party.append(
                measurements_per_party.reshape(
                    (-1, O_card, self._nr_properties)))
        self._ortho_groups = list(chain.from_iterable(self._ortho_groups_per_party))

        #New - orthogonal indices. To be used to accelerate monomial instantiation.
        offset = 0
        self._ortho_idxs_per_party = []
        self._ortho_idxs = []
        for ortho_groups_of_party in self._ortho_groups_per_party:
            ortho_idxs_of_party = []
            for ortho_group in ortho_groups_of_party:
                l = len(ortho_group)
                block = np.arange(l)
                block += offset
                ortho_idxs_of_party.append(block)
                offset += l
            self._ortho_idxs_per_party.append(np.array(ortho_idxs_of_party))
            self._ortho_idxs.extend(ortho_idxs_of_party)
        self._template_idxs = np.array([ortho_idx_group[0] for ortho_idx_group in self._ortho_idxs], dtype=int)

        self._lexorder = np.vstack(self._ortho_groups).astype(self._np_dtype)
        self.party_from_lexidx = self._lexorder[:,0]
        self.party_from_templateidx = self.party_from_lexidx[self._template_idxs]
        self._nr_operators = len(self._lexorder)

        self._lexorder_for_factorization = np.array([
            self._inflation_indices_hash[op.tobytes()]
            for op in self._lexorder[:, 1:-2]],
            dtype=np.intc)

        # Set up compatible measurements
        if self._nonclassical_sources.any():
            self._default_notcomm = commutation_matrix(self._lexorder,
                                                       self.sources_to_check_for_party_pair_commutation,
                                                       False)
        else:
            self._default_notcomm = np.zeros(
                (self._lexorder.shape[0], self._lexorder.shape[0]),
                dtype=bool)

        _lexorder_to_original = self.rectify_fake_setting(
            self._lexorder[:, [0, -2, -1]])
        (self.original_dag_events,
         self._canonical_lexids,
         self._lexidx_to_origidx) = np.unique(_lexorder_to_original,
                                              return_index=True,
                                              return_inverse=True, axis=0)

        # Symmetries implied by the inflation
        self.symmetries = self.inflation_symmetries


    @property
    def _compatible_template_measurements(self):
        return np.invert(self._default_notcomm[np.ix_(self._template_idxs, self._template_idxs)])

    def all_and_maximal_compatible_templates(self,
                                             max_n=0,
                                             isolate_maximal=True):
        return all_and_maximal_cliques(self._compatible_template_measurements,
                                       max_n=max_n,
                                       isolate_maximal=isolate_maximal)


    def __repr__(self):
        if self._classical_sources.all():
            source_info = "All sources are classical."
        elif self._nonclassical_sources.all():
            source_info = "All sources are quantum."
        else:
            classical_sources = self._actual_sources[self._classical_sources.astype(bool)]
            quantum_sources   = self._actual_sources[self._nonclassical_sources.astype(bool)]
            if len(classical_sources) > 1:
                extra_1 = "s"
                extra_2 = "are"
            else:
                extra_1 = ""
                extra_2 = "is"
            source_info = f"Source{extra_1} " + ", ".join(classical_sources) \
                + f" {extra_2} classical, and "
            if len(quantum_sources) > 1:
                extra_1 = "s"
                extra_2 = "are"
            else:
                extra_1 = ""
                extra_2 = "is"
            source_info += f"source{extra_1} " + ", ".join(quantum_sources) \
                + f" {extra_2} quantum."
        return ("InflationProblem with " + str(self.dag) +
                " as DAG, " + str(self.outcomes_per_party) +
                " outcomes per party, " + str(self.settings_per_party) +
                " settings per party and " +
                str(self.inflation_level_per_source) +
                " inflation copies per source. " + source_info)

    ###########################################################################
    # HELPER UTILITY FUNCTION                                                 #
    ###########################################################################
    @cached_property
    def _lexorder_lookup(self) -> dict:
        """Creates helper dictionary for quick lookup of lexorder indices.

        Returns
        -------
        dict
            Mapping an operator in .tobytes() for quick lookup of its index
            in the lexorder.
        """
        return {op.tobytes(): i for i, op in enumerate(self._lexorder)}
    
    def mon_to_lexrepr(self, mon: np.ndarray) -> np.ndarray:
        ops_as_hashes = list(map(self._from_2dndarray, mon))
        try:
            return np.array(partsextractor(self._lexorder_lookup, ops_as_hashes), dtype=np.intc)
        except KeyError:
            raise Exception(f"Failed to interpret\n{mon}\n relative to specified lexorder \n{self._lexorder}.")

    def _from_2dndarray(self, array2d: np.ndarray) -> bytes:
        """Obtains the bytes representation of an array. The library uses this
        representation as hashes for the corresponding monomials.

        Parameters
        ----------
        array2d : numpy.ndarray
            Monomial encoded as a 2D array.
        """
        return np.asarray(array2d, dtype=self._np_dtype).tobytes()

    @cached_property
    def _any_inflation(self) -> bool:
        return np.any(self.inflation_level_per_source > 1)

    @cached_property
    def _lexrepr_to_dicts(self) -> np.ndarray:
        """Map 1D array lexorder encoding of a monomial, to a string representation.

        Returns
        -------
        numpy.ndarray
            Array of the same length as lexorder, where the i-th element is the
            dictionary interpretation of the i-th operator in the lexorder.
        """
        return np.array([self._interpret_operator(op) for op in self._lexorder])

    @cached_property
    def _lexrepr_to_names(self) -> np.ndarray:
        """Map 1D array lexorder encoding of a monomial, to a string representation.

        Returns
        -------
        numpy.ndarray
            Array of the same length as lexorder, where the i-th element is the
            string representation of the i-th operator in the lexorder.
        """
        return np.asarray([self._interpretation_to_name(
            op_dict,
            include_copy_indices=self._any_inflation)
                for op_dict in self._lexrepr_to_dicts.flat], dtype=object)

    @cached_property
    def _original_event_names(self) -> np.ndarray:
        """Map 1D array lexorder-like encoding of an event, to a string representation.

        Returns
        -------
        numpy.ndarray
            Array of the same length as original events, where the i-th element is the
            string representation of the i-th operator in the event list.
        """
        return np.asarray([self._interpretation_to_name(
            self._interpret_operator(event),
            include_copy_indices=False)
                for event in self.original_dag_events], dtype=object)

    @cached_property
    def _lexrepr_to_copy_index_free_names(self) -> np.ndarray:
        """Map 1D array lexorder encoding of a monomial, to a string representation.

        Returns
        -------
        numpy.ndarray
            Array of the same length as lexorder, where the i-th element is the
            string representation of the i-th operator in the lexorder.
        """
        if not self._any_inflation:
            return self._lexrepr_to_names
        else:
            return np.asarray([self._interpretation_to_name(
                op_dict,
                include_copy_indices=False)
                for op_dict in self._lexrepr_to_dicts.flat], dtype=object)

    @cached_property
    def _lexrepr_to_all_names(self) -> np.ndarray:
        """Map 1D array lexorder encoding of a monomial to
        a variety of possible string representations.

        Returns
        -------
        numpy.ndarray
            Array of the same length as lexorder, where the i-th element is a
            collection of string representations of the i-th operator in the
            lexorder.
        """
        old_names_v1 = [self._interpretation_to_name_old(op_dict,
                                                         replace_trivial_setting=True)
                        for op_dict in self._lexrepr_to_dicts.flat]
        old_names_v2 = [legacy_name.replace('∅','0') for legacy_name in old_names_v1]
        return np.stack((
            self._lexrepr_to_names,
            self._lexrepr_to_copy_index_free_names,
            old_names_v1,
            old_names_v2
        ), 1).astype(object)

    @cached_property
    def _lexrepr_to_symbols(self) -> np.ndarray:
        """For each operator in the lexorder, create a sympy symbol with the 
        same name as returned by InflationProblem._lexrepr_to_names()

        Returns
        -------
        list
            List of the same length as lexorder, where the i-th element is the
            string representation of the i-th operator in the lexorder.
        """
        return np.array([Symbol(name, commutative=False)
                         for name in self._lexrepr_to_names],
                        dtype=object)

    @cached_property
    def names_to_ints(self) -> dict:
        """Map party names to integers denoting their position in the list of
        parties, as specified by 'order', offset by 1 (i.e., first party is 1).

        Used for the 2D array representation of monomials.

        Returns
        -------
        dict
            Dictionary mapping party names to integers.
        """
        return {name: i + 1 for i, name in enumerate(self.names)}

    def _1d_to_symbol(self, repr_1d: np.ndarray) -> Symbol:
        """Convert a 1D representation of an operator to a sympy symbol.
        
        Parameters
        ----------
        repr_1d : np.ndarray
            1D representation of an operator.

        Returns
        -------
        sympy.Symbol
            Sympy symbol representing the operator.
        """
        name = self._interpretation_to_name(self._interpret_operator(repr_1d))
        if not self._any_inflation:
            # If there is no inflation, we can remove the copy indices
            name = name.split('^{')[0] + name.split('}')[1]
        return Symbol(name, commutative=False)

    ###########################################################################
    # FUNCTIONS PERTAINING TO KNOWABILITY                                     #
    ###########################################################################
    def _interpret_operator(self, op: np.ndarray) -> dict:
        interpretation = {}
        party_int = op[0] -1
        interpretation["Party as Integer"] = party_int
        interpretation["Party"] = self.names[party_int]
        outcome_int = op[-1]
        interpretation["Outcome"] = outcome_int
        setting_as_single_int = op[-2]
        interpretation["Composite Setting"] = setting_as_single_int
        interpretation["Composite Setting is Trivial"] = (self.settings_per_party[party_int] == 1)
        setting_as_tuple = self.effective_to_parent_settings[party_int][setting_as_single_int]
        interpretation["Setting as Tuple"] = setting_as_tuple
        private_setting = setting_as_tuple[0]
        interpretation["Private Setting"] = private_setting
        interpretation["Private Setting is Trivial"] = (self.private_settings_per_party[party_int] == 1)
        outcomes_of_parents = setting_as_tuple[1:]
        parents_in_play_as_ints = self.parents_per_party[party_int]
        interpretation["Parents in-play as Integers"] = parents_in_play_as_ints
        parents_in_play_as_names = partsextractor(self.names, parents_in_play_as_ints)
        non_private_setting_dict = dict(zip(parents_in_play_as_names, outcomes_of_parents))
        interpretation["Do Values"] = non_private_setting_dict
        if len(op)==3:
            return interpretation
        interpretation["Copy Indices"] = op[1:-2]
        relevant_slots = np.logical_and(
            self.inflation_level_per_source > 0,
            interpretation["Copy Indices"] > 0
        )
        interpretation["Relevant Copy Indices"] = interpretation["Copy Indices"][relevant_slots]
        return interpretation

    @staticmethod
    def _make_interpretation_hashable(op_as_dict):
        return (int(op_as_dict["Party as Integer"]),
                tuple(int(i) for i in op_as_dict["Copy Indices"]),
                tuple(int(i) for i in op_as_dict["Setting as Tuple"]),
                int(op_as_dict["Outcome"]))

    @staticmethod
    def _interpretation_to_name(op: dict, include_copy_indices=True) -> str:
        op_as_str = op["Party"]
        if not op["Private Setting is Trivial"]:
            op_as_str += '_'+str(op["Private Setting"])
        if include_copy_indices:
            if len(op["Relevant Copy Indices"]):
                copy_index_string = '^{'
                copy_index_string += ','.join(map(str,op["Relevant Copy Indices"].flat))
                copy_index_string += '}'
                op_as_str += copy_index_string
            # copy_indices_string = "_" + "_".join(map(str, op["Copy Indices"]))
            # op_as_str += copy_indices_string
        op_as_str += f"={op['Outcome']}"
        if len(op["Do Values"]):
            do_values_string = "|do("
            do_values_string += ','.join((f"{p}={v}" for p, v in op["Do Values"].items()))
            do_values_string += ')'
            op_as_str += do_values_string
        return op_as_str

    @staticmethod
    def _interpretation_to_name_old(op: dict, replace_trivial_setting=True) -> str:
        op_as_str = op["Party"]
        copy_indices_string = "_" + "_".join(map(str, op["Copy Indices"]))
        op_as_str += copy_indices_string
        if replace_trivial_setting and op["Composite Setting is Trivial"]:
            op_as_str += "_∅"
        else:
            op_as_str += "_" + str(op["Composite Setting"])
        op_as_str += "_" + str(op['Outcome'])
        return op_as_str

    def _is_knowable_q_non_networks(self, monomial: np.ndarray) -> bool:
        """Checks if a monomial (written as a sequence of operators in 2d array
        form) corresponds to a knowable probability. The function assumes that
        the candidate monomial already passed the preliminary knowable test
        from `inflation.sdp.quantum_tools.py`.
        If the scenario is a network, this function always returns ``True``.

        Parameters
        ----------
        monomial : numpy.ndarray
            An internal representation of a monomial as a 2d numpy array. Each
            row in the array corresponds to an operator. For each row, the
            zeroth element represents the party, the last element represents
            the outcome, the second-to-last element represents the setting, and
            the remaining elements represent inflation copies.

        Returns
        -------
        bool
            Whether the monomial can be assigned a knowable probability.
        """
        # Parties start at 1 in our notation
        parties_in_play = np.asarray(monomial)[:, 0] - 1
        parties_absent_bitvector    = np.ones(self.nr_parties, dtype=bool)
        parties_absent_bitvector[parties_in_play] = False
        for p in parties_in_play.flat:
            if parties_absent_bitvector[self.parents_per_party[p]].any():
                return False # Case of not an ancestrally closed set
        # Parties start at 1 in our notation
        outcomes_by_party = {(o[0] - 1): o[-1] for o in monomial}
        for o in monomial:
            party_index           = o[0] - 1
            effective_setting     = o[-2]
            o_nonprivate_settings = \
                self.effective_to_parent_settings[
                                  party_index][effective_setting][1:]
            outcomes_of_parent_parties = partsextractor(outcomes_by_party, self.parents_per_party[party_index])
            if not np.array_equal(outcomes_of_parent_parties, o_nonprivate_settings):
                return False
        return True

    def rectify_fake_setting(self, monomial: np.ndarray) -> np.ndarray:
        """When constructing the monomials in a non-network scenario, we rely
        on an internal representation of operators where the integer denoting
        the setting actually is an 'effective setting' that encodes, in
        addition to the 'private setting' that each party is free to choose,
        the values of all the parents of the variable in question, which also
        effectively act as settings. This function resets this 'effective
        setting' integer to the true 'private setting' integer. It is useful to
        relate knowable monomials to their meaning as conditional events in
        non-network scenarios. If the scenario is a network, this function does
        nothing.

        Parameters
        ----------
            monomial : numpy.ndarray
                An internal representation of a monomial as a 2d numpy array.
                Each row in the array corresponds to an operator. For each row,
                the zeroth element represents the party, the last element
                represents the outcome, the second-to-last element represents
                the setting, and the remaining elements represent inflation
                copies.

        Returns
        -------
        numpy.ndarray
            The monomial with the index of the setting representing only the
            private setting.
        """
        new_mon = np.array(monomial, copy=False)
        for o in new_mon:
            party_index        = o[0] - 1  # Parties start at 1 our notation
            effective_setting  = o[-2]
            o_private_settings = \
                self.effective_to_parent_settings[
                                   party_index][effective_setting][0]
            o[-2] = o_private_settings
        return new_mon

    ###########################################################################
    # FUNCTIONS PERTAINING TO FACTORIZATION                                   #
    ###########################################################################
    def factorize_monomial_2d(self,
                              monomial_as_2darray: np.ndarray,
                              canonical_order=False) -> Tuple[np.ndarray, ...]:
        """Split a moment/expectation value into products of moments according
        to the support of the operators within the moment. The moment is
        encoded as a 2d array where each row is an operator. If
        ``monomial=A*B*C*B`` then row 1 is ``A``, row 2 is ``B``, row 3 is
        ``C``, and row 4 is ``B``. In each row, the columns encode the
        following information:

          * First column: The party index, *starting from 1* (e.g., 1 for
            ``A``, 2 for ``B``, etc.)
          * Last two columns: The input ``x`` starting from zero, and then the
            output ``a`` starting from zero.
          * In between: This encodes the support of the operator. There are as
            many columns as sources/quantum states. Column `j` represents
            source `j-1` (-1 because the first column represents the party). If
            the value is 0, then this operator does not measure this source. If
            the value is, e.g., 2, then this operator is acting on copy 2 of
            source `j-1`.

        The output is a tuple of ndarrays where each array represents another
        monomial s.t. their product is equal to the original monomial.

        Parameters
        ----------
        monomial_as_2darray : numpy.ndarray
            Monomial in 2d array form.
        canonical_order: bool, optional
            Whether to return the different factors in a canonical order.

        Returns
        -------
        Tuple[numpy.ndarray]
            A tuple of ndarrays, where each array represents an atomic monomial
            factor.
        Examples
        --------
        >>> monomial = np.array([[1, 0, 1, 1, 0, 0],
                                 [2, 1, 0, 2, 0, 0],
                                 [1, 0, 3, 3, 0, 0],
                                 [3, 3, 5, 0, 0, 0],
                                 [3, 1, 4, 0, 0, 0],
                                 [3, 6, 6, 0, 0, 0],
                                 [3, 4, 5, 0, 0, 0]])
        >>> factorised = factorize_monomial(monomial_as_2darray)
        [array([[1, 0, 1, 1, 0, 0]]),
         array([[1, 0, 3, 3, 0, 0]]),
         array([[2, 1, 0, 2, 0, 0],
                [3, 1, 4, 0, 0, 0]]),
         array([[3, 3, 5, 0, 0, 0],
                [3, 4, 5, 0, 0, 0]]),
         array([[3, 6, 6, 0, 0, 0]])]
        """
        lexmon_factors = self.factorize_monomial_1d(self.mon_to_lexrepr(monomial_as_2darray),
                                          canonical_order=canonical_order)
        return tuple(self._lexorder[lexmon] for lexmon in lexmon_factors)

    def factorize_monomial_1d(self,
                              monomial_as_1darray: np.ndarray,
                              canonical_order=False) -> Tuple[np.ndarray, ...]:
        """Split a moment/expectation value into products of moments according
        to the support of the operators within the moment. The moment is
        encoded as a 1d array representing elements of the lexorder.

        The output is a tuple of ndarrays where each array represents another
        monomial s.t. their product is equal to the original monomial.

        Parameters
        ----------
        monomial_as_1darray : numpy.ndarray
            Monomial in 1d array form.
        canonical_order: bool, optional
            Whether to return the different factors in a canonical order.

        Returns
        -------
        Tuple[numpy.ndarray]
            A tuple of ndarrays, where each array represents a picklist for an atomic monomial
            factor.
        """
        if (not self.ever_factorizes) or len(monomial_as_1darray) <= 1:
            return (monomial_as_1darray,)

        inflation_indices_position = self._lexorder_for_factorization[
            monomial_as_1darray]
        unique_inflation_indices_positions, reversion_key = np.unique(inflation_indices_position, return_inverse=True)
        adj_mat = self._inflation_indices_overlap[np.ix_(unique_inflation_indices_positions, unique_inflation_indices_positions)]
        component_labels = nb_classify_disconnected_components(adj_mat)
        nof_components = component_labels.max(initial=0) + 1
        disconnected_components = tuple(monomial_as_1darray[np.take(component_labels == i, reversion_key)]
            for i in range(nof_components))

        if canonical_order:
            disconnected_components = tuple(sorted(disconnected_components,
                                                   key=lambda x: x.tobytes()))
        return disconnected_components

    ###########################################################################
    # FUNCTIONS PERTAINING TO SYMMETRIES                                      #
    ###########################################################################
    @cached_property
    def inflation_symmetries(self) -> np.ndarray:
        """Calculates all the symmetries pertaining to the set of generating
        monomials due to copy index relabelling. The new set of operators is a
        permutation of the old. The function outputs a list of all permutations.

        Returns
        -------
        numpy.ndarray[int]
            The permutations of the lexicographic order implied by the inflation
            symmetries.
        """
        sources_with_copies = [source for source, inf_level
                               in enumerate(self.inflation_level_per_source)
                               if inf_level > 1]
        if len(sources_with_copies):
            permutation_failed = False
            symmetries         = []
            identity_perm      = np.arange(self._nr_operators, dtype=np.intc)
            for source in tqdm(sources_with_copies,
                               disable=not self.verbose,
                               desc="Calculating symmetries   ",
                               leave=True,
                               position=0):
                one_source_symmetries = [identity_perm]
                inf_level = self.inflation_level_per_source[source]
                perms = format_permutations(list(
                    permutations(range(inf_level)))[1:])
                for permutation in perms:
                    adjusted_ops = apply_source_perm(self._lexorder,
                                                     source,
                                                     permutation)
                    try:
                        new_order = np.fromiter(
                            (self._lexorder_lookup[op.tobytes()]
                             for op in adjusted_ops),
                            dtype=np.intc
                        )
                        one_source_symmetries.append(new_order)
                    except KeyError:
                        permutation_failed = True
                symmetries.append(np.asarray(one_source_symmetries,
                                             dtype=np.intc))
            if permutation_failed and (self.verbose > 0):
                warn("The generating set is not closed under source swaps."
                     + " Some symmetries will not be implemented.")
            return reduce(perm_combiner, symmetries)
        return np.arange(self._nr_operators, dtype=np.intc)[np.newaxis]

    def add_symmetries(self,
                       new_symmetries: Union[np.ndarray, List[np.ndarray]]
                       ):
        """Adds new symmetries, represented as permutations of the
        lexicographical order, to the list of symmetries in the scenario.

        Parameters
        -------
        new_symmetries : numpy.ndarray[int], List[np.ndarray[int]]
            The permutations of the lexicographic order that conform the
            symmetries to be included.
        """
        self.symmetries = group_elements_from_generators(
            np.vstack((self.symmetries, new_symmetries)))

    def reset_symmetries(self):
        """Remove all the symmetries of the scenario, keeping only the ones
            that are implied by the inflation structure.        
        """
        self.symmetries = self.inflation_symmetries

    @cached_property
    def _lexorder_hashable_interpretation_decoder(self):
        return {self._make_interpretation_hashable(op_as_dict): i for
                    i, op_as_dict in enumerate(self._lexrepr_to_dicts)}

    @cached_property
    def _party_relabelling_symmetries(self) -> np.ndarray:
        """Return a list of all party relabelling symmetries (each proceeded by
        its associated source relabelling symmetry) consistent with the
        graphical symmetries of the original DAG, subject to matching
        cardinalities of inputs and outputs for all exchanged parties, and
        subject to matching inflation levels for all exchanged sources.

        Returns
        -------
        Tuple[numpy.ndarray, numpy.ndarray]
            The first array gives permutations of the lexorder, the second array
            gives the permutations of the original events.
        """
        nr_sources = self.nr_sources
        g1 = nx.DiGraph()
        g1.add_nodes_from(range(self.nr_parties+nr_sources))
        for s, bool_children in enumerate(self.hypergraph):
            for child in np.flatnonzero(bool_children):
                g1.add_edge(s, child+self.nr_sources)
        for c, parents in enumerate(self.parents_per_party):
            for p in parents:
                g1.add_edge(p+self.nr_sources, c + nr_sources)
        GMgen = isomorphism.DiGraphMatcher(g1, g1)
        discovered_automorphisms = list()
        for mapping in GMgen.isomorphisms_iter():
            valid_automorphism = True
            for s, inf_level in enumerate(
                    self.inflation_level_per_source.flat):
                if not valid_automorphism:
                    break
                new_s = mapping[s]
                valid_automorphism = (
                        self.inflation_level_per_source[new_s] == inf_level)
            for p, (card_in, card_out) in enumerate(zip(
                    self.settings_per_party.flat,
                    self.outcomes_per_party.flat)):
                if not valid_automorphism:
                    break
                new_p = mapping[p + self.nr_sources] - nr_sources
                valid_automorphism = (
                        (self.settings_per_party[new_p] == card_in)
                    and (self.outcomes_per_party[new_p] == card_out))
            if valid_automorphism:
                discovered_automorphisms.append((
                    np.fromiter((mapping[i] for i in
                                 range(nr_sources)),
                                dtype=int),
                     np.fromiter((mapping[i + nr_sources] - nr_sources
                                  for i in
                                  range(self.nr_parties)),
                                 dtype=int),
                     ))

        lexorder_perms = []
        for automorphism in discovered_automorphisms:
            source_perm, party_perm = automorphism
            template = self._lexorder.copy()
            for p, new_p in zip(range(self.nr_parties), party_perm):
                template[self._lexorder[:, 0] == p + 1, 0] = new_p + 1
            new_source_perm = np.argsort(source_perm)
            template = template[:, [0]+(1+new_source_perm).tolist() + [-2, -1]]
            lexorder_perm = np.array([self._lexorder_lookup[op.tobytes()]
                                      for op in template])
            lexorder_perms += [lexorder_perm]
        return np.array(lexorder_perms)

    @cached_property
    def _setting_specific_outcome_relabelling_symmetries(self) -> np.ndarray:
        """
        Yields all possible setting relabellings paired with all possible
        setting-dependant outcome relabellings as permutations of the events on
        the original graph. Seperated by party, so that iteration will involve
        itertools.product.
        """
        identity_perm = np.arange(self._nr_operators, dtype=int)
        sym_generators = [identity_perm]

        for p in range(self.nr_parties):
            # We do not attempt outcome relabelling on parties with children
            # and parents!
            if self.has_children[p] and (self.settings_per_party[p] > 1):
                break
            for x in range(self.private_settings_per_party[p]):
                for i, perm in enumerate(
                                   permutations(
                                       range(self.outcomes_per_party[p]))):
                    if i == 0:
                        continue  # skip empty perm
                    new_interpretations = [op_as_dict.copy()
                                           for op_as_dict
                                           in self._lexrepr_to_dicts]
                    lexorder_perm = []
                    for op_as_dict in new_interpretations:
                        if p in op_as_dict["Parents in-play as Integers"]:
                            to_adjust = list(op_as_dict["Setting as Tuple"])
                            old_value = to_adjust[p+1]
                            new_value = perm[old_value]
                            to_adjust[p+1] = new_value
                            op_as_dict["Setting as Tuple"] = tuple(to_adjust)
                        if ((op_as_dict["Party as Integer"] == p)
                            and (op_as_dict["Private Setting"] == x)):
                            old_value = op_as_dict["Outcome"]
                            op_as_dict["Outcome"] = perm[old_value]
                        lexorder_perm.append(
                            self._lexorder_hashable_interpretation_decoder[
                                self._make_interpretation_hashable(op_as_dict)])
                    lexorder_perm = np.array(lexorder_perm)
                    sym_generators.append(lexorder_perm)
        return np.array(sym_generators)

    @cached_property
    def _party_specific_setting_relabelling_symmetries(self) -> np.ndarray:
        """
        Yields all possible setting relabellings that are specific to just one
        party as permutations of the events on the original graph. Seperated by
        party, so that iteration will involve itertools.product.
        """
        identity_perm = np.arange(self._nr_operators, dtype=int)
        sym_generators = [identity_perm]
        for p in range(self.nr_parties):
            # Since we are only adjusting PRIVATE setting, we can proceed even
            # if the party has children
            for i, perm in enumerate(
                            permutations(
                                range(self.private_settings_per_party[p]))):
                if i == 0:
                    continue  # skip empty perm
                new_interpretations = [op_as_dict.copy() for op_as_dict in
                                       self._lexrepr_to_dicts]
                lexorder_perm = []
                for op_as_dict in new_interpretations:
                    if op_as_dict["Party as Integer"] == p:
                        to_adjust = list(op_as_dict["Setting as Tuple"])
                        old_value = to_adjust[0]
                        new_value = perm[old_value]
                        to_adjust[0] = new_value
                        op_as_dict["Setting as Tuple"] = tuple(to_adjust)
                    lexorder_perm.append(
                        self._lexorder_hashable_interpretation_decoder[
                            self._make_interpretation_hashable(op_as_dict)])
                lexorder_perm = np.array(lexorder_perm)
                sym_generators.append(lexorder_perm)
        return np.array(sym_generators)

    @cached_property
    def _all_possible_symmetries(self) -> np.ndarray:
        group_generators = np.vstack((
            self._party_relabelling_symmetries,
            self._party_specific_setting_relabelling_symmetries,
            self._setting_specific_outcome_relabelling_symmetries))
        group_elements = group_elements_from_generators(group_generators)
        return group_elements



import unittest
import numpy as np
import warnings
from itertools import product

from inflation import InflationProblem, InflationSDP, InflationLP

bilocalDAG = {"h1": ["A", "B"], "h2": ["B", "C"]}
bilocality = InflationProblem(dag=bilocalDAG,
                              settings_per_party=[1, 1, 1],
                              outcomes_per_party=[2, 2, 2],
                              inflation_level_per_source=[2, 2])
bilocality_c = InflationProblem(dag=bilocalDAG,
                                settings_per_party=[1, 1, 1],
                                outcomes_per_party=[2, 2, 2],
                                inflation_level_per_source=[2, 2],
                                classical_sources="all")
bilocalSDP = InflationSDP(bilocality)

trivial = InflationProblem({"h": ["v"]},
                           outcomes_per_party=[2],
                           settings_per_party=[2],
                           inflation_level_per_source=[2]
                           )
trivial_c = InflationProblem({"h": ["v"]},
                           outcomes_per_party=[2],
                           settings_per_party=[2],
                           inflation_level_per_source=[2],
                           classical_sources="all"
                           )

bellScenario = InflationProblem({"Lambda": ["A", "B"]},
                                outcomes_per_party=[2, 2],
                                settings_per_party=[2, 2],
                                inflation_level_per_source=[1])
bellScenario_c = InflationProblem({"Lambda": ["A", "B"]},
                                    outcomes_per_party=[2, 2],
                                    settings_per_party=[2, 2],
                                    inflation_level_per_source=[1],
                                    classical_sources='all')

class TestMonomialGeneration(unittest.TestCase):
    bilocalSDP_commuting = InflationSDP(bilocality_c)
    # Column structure for the NPA level 2 in a tripartite scenario
    col_structure = [[],
                     [0], [1], [2],
                     [0, 0], [0, 1], [0, 2], [1, 1], [1, 2], [2, 2]]
    # Monomials for the NPA level 2 in the bilocality scenario
    meas = bilocality_c.measurements_symbolic
    A_1_0_0_0 = meas[0][0][0][0]
    A_2_0_0_0 = meas[0][1][0][0]
    B_1_1_0_0 = meas[1][0][0][0]
    B_1_2_0_0 = meas[1][1][0][0]
    B_2_1_0_0 = meas[1][2][0][0]
    B_2_2_0_0 = meas[1][3][0][0]
    C_0_1_0_0 = meas[2][0][0][0]
    C_0_2_0_0 = meas[2][1][0][0]
    cols = [1, A_1_0_0_0, A_2_0_0_0, B_1_1_0_0, B_1_2_0_0, B_2_1_0_0,
            B_2_2_0_0, C_0_1_0_0, C_0_2_0_0, A_1_0_0_0*A_2_0_0_0,
            A_1_0_0_0*B_1_1_0_0, A_1_0_0_0*B_1_2_0_0, A_1_0_0_0*B_2_1_0_0,
            A_1_0_0_0*B_2_2_0_0, A_2_0_0_0*B_1_1_0_0, A_2_0_0_0*B_1_2_0_0,
            A_2_0_0_0*B_2_1_0_0, A_2_0_0_0*B_2_2_0_0, A_1_0_0_0*C_0_1_0_0,
            A_1_0_0_0*C_0_2_0_0, A_2_0_0_0*C_0_1_0_0, A_2_0_0_0*C_0_2_0_0,
            B_1_1_0_0*B_1_2_0_0, B_1_1_0_0*B_2_1_0_0, B_1_1_0_0*B_2_2_0_0,
            B_1_2_0_0*B_1_1_0_0, B_1_2_0_0*B_2_1_0_0, B_1_2_0_0*B_2_2_0_0,
            B_2_1_0_0*B_1_1_0_0, B_2_1_0_0*B_2_2_0_0, B_2_2_0_0*B_1_2_0_0,
            B_2_2_0_0*B_2_1_0_0, B_1_1_0_0*C_0_1_0_0, B_1_1_0_0*C_0_2_0_0,
            B_1_2_0_0*C_0_1_0_0, B_1_2_0_0*C_0_2_0_0, B_2_1_0_0*C_0_1_0_0,
            B_2_1_0_0*C_0_2_0_0, B_2_2_0_0*C_0_1_0_0, B_2_2_0_0*C_0_2_0_0,
            C_0_1_0_0*C_0_2_0_0]
    actual_cols = []
    for col in cols:
        actual_cols.append(bilocalSDP.mon_to_lexrepr(bilocalSDP._interpret_name(col)))
    actual_cols_2d = [bilocalSDP._lexorder[lexmon] for lexmon in actual_cols]

    def test_generating_columns_c(self):
        truth = 37
        columns = self.bilocalSDP_commuting.build_columns(self.col_structure)
        self.assertEqual(len(columns), truth,
                         "With commuting variables, there are  " +
                         str(len(columns)) + " columns but " + str(truth) +
                         " were expected.")

    def test_generating_columns_nc(self):
        truth = 41
        columns = bilocalSDP.build_columns(self.col_structure)
        self.assertEqual(len(columns), truth,
                         "With noncommuting variables, there are  " +
                         str(len(columns)) + " columns but " + str(truth) +
                         " were expected.")

    def test_generation_from_columns(self):
        columns = bilocalSDP.build_columns(self.actual_cols_2d)
        areequal = all(np.array_equal(r[0].T, np.array(r[1]).T)
                       for r in zip(columns, self.actual_cols))
        self.assertTrue(areequal, "The direct copying of columns is failing.")

    def test_generation_from_lol(self):
        columns = bilocalSDP.build_columns(self.col_structure)
        areequal = all(np.array_equal(r[0].T, np.array(r[1]).T)
                       for r in zip(columns, self.actual_cols))
        self.assertTrue(areequal,
                        "Parsing a list-of-list description of columns fails.")

    def test_generation_from_str(self):
        columns = bilocalSDP.build_columns("npa2")
        areequal = all(np.array_equal(r[0].T, np.array(r[1]).T)
                       for r in zip(columns, self.actual_cols))
        self.assertTrue(areequal,
                        "Parsing NPA levels with string description fails.")
        columns = bilocalSDP.build_columns("local2", max_monomial_length=2)
        diff = set(tuple(col) for col in columns
                   ).difference(
                       set(tuple(col)
                           for col in self.actual_cols))
        self.assertTrue(len(diff) == 0,
                        "Parsing local levels with string description fails.")
        columns = bilocalSDP.build_columns("local221", max_monomial_length=2)
        diff = set(tuple(col) for col in columns
                   ).difference(
                       set(tuple(col)
                           for col in self.actual_cols[:-1]))
        self.assertTrue(len(diff) == 0,
                        "Parsing local levels with individual string " +
                        "descriptions fails.")
        columns = bilocalSDP.build_columns("physical2",
                                           max_monomial_length=2)
        physical = (self.actual_cols[:22] + [self.actual_cols[24]]
                    + [self.actual_cols[26]] + self.actual_cols[32:])
        diff = set(tuple(col) for col in columns
                   ).difference(
                       set(tuple(col)
                           for col in physical))
        self.assertTrue(len(diff) == 0,
                        "Parsing physical levels with global string " +
                        "description fails.")
        columns = bilocalSDP.build_columns("physical", max_monomial_length=2)
        diff = set(tuple(col) for col in columns
                   ).difference(
                       set(tuple(col)
                           for col in physical))
        self.assertTrue(len(diff) == 0,
                        "Parsing physical levels without further " +
                        "description fails.")
        columns = bilocalSDP.build_columns("physical121",
                                           max_monomial_length=2)
        diff = set(tuple(col) for col in columns
                   ).difference(
                       set(tuple(col)
                           for col in (self.actual_cols[:9]
                                       + self.actual_cols[10:22]
                                       + [self.actual_cols[24]]
                                       + [self.actual_cols[26]]
                                       + self.actual_cols[32:-1])))
        self.assertTrue(len(diff) == 0,
                        "Parsing physical levels with individual string " +
                        "descriptions fails.")

    def test_generation_with_identities(self):
        oneParty = InflationSDP(InflationProblem({"h": ["v"]}, [2], [2], [1]))
        columns  = oneParty.build_columns([[], [0, 0]])
        truth    = [np.array([[1, 1, 0, 0], [1, 1, 1, 0]]),
                    np.array([[1, 1, 1, 0], [1, 1, 0, 0]])]
        truth    = [np.empty((0,4), dtype=int)] + [oneParty.mon_to_lexrepr(mon) 
                                                   for mon in truth]
        self.assertTrue(len(columns) == len(truth),
                        "Generating columns with identities is not producing "
                        + "the correct number of columns.")
        areequal = all(np.array_equiv(r[0].T, np.array(r[1]).T)
                       for r in zip(columns, truth))
        self.assertTrue(areequal,
                        "The column generation is not capable of handling " +
                        "monomials that reduce to the identity" +
                        f"\ngot: {columns} \nexpected: {truth}")


class TestReset(unittest.TestCase):
    sdp = InflationSDP(trivial)
    sdp.generate_relaxation("npa1")
    physical_bounds = {m: 0. for m in sdp.hermitian_moments}
    del physical_bounds[sdp.One]

    def prepare_objects(self, infSDP):
        var1 = trivial.measurements_symbolic[0][0][0][0]
        var2 = trivial.measurements_symbolic[0][0][1][0]
        infSDP.set_objective(var1, "max")
        infSDP.set_bounds({var1*var2: 0.9}, "up")
        infSDP.set_bounds({var1*var2: 0.1}, "lo")
        infSDP.set_values({var2: 0.5})

    def test_reset_all(self):
        self.prepare_objects(self.sdp)
        self.sdp.reset("all")
        self.assertEqual(self.sdp.moment_lowerbounds,
                         self.physical_bounds,
                         "Resetting lower bounds fails.")
        self.assertEqual(self.sdp.moment_upperbounds, {},
                         "Resetting processed upper bounds fails.")
        self.assertEqual(self.sdp.objective, {self.sdp.One: 0.},
                         "Resetting the objective function fails.")
        self.assertEqual(self.sdp.semiknown_moments, {},
                         "Resetting the known values fails to empty " +
                         "semiknown_moments.")
        self.assertEqual(self.sdp.known_moments, {self.sdp.One: 1.},
                         "Resetting the known values fails to empty " +
                         "known_moments.")

    def test_reset_bounds(self):
        self.prepare_objects(self.sdp)
        correct = {key: val for key, val in self.physical_bounds.items()
                   if key not in self.sdp.known_moments}
        self.sdp.reset("bounds")
        self.assertEqual(self.sdp.moment_lowerbounds,
                         correct,
                         "Resetting lower bounds fails.")
        self.assertEqual(self.sdp.moment_upperbounds, {},
                         "Resetting upper bounds fails.")
        self.assertTrue(len(self.sdp.objective) == 2,
                        "Resetting the bounds resets the objective function.")
        self.assertTrue(len(self.sdp.known_moments) == 3,
                        "Resetting the bounds resets the known moments.")

    def test_reset_some(self):
        self.prepare_objects(self.sdp)
        self.sdp.reset(["objective", "values"])
        self.assertEqual(self.sdp.objective, {self.sdp.One: 0.},
                         "Resetting the objective function fails.")
        self.assertEqual(self.sdp.semiknown_moments, {},
                        "Resetting the known values fails to empty " +
                        "semiknown_moments.")
        self.assertEqual(self.sdp.known_moments, {self.sdp.One: 1.},
                         "Resetting the known values fails to empty " +
                         "known_moments.")
        self.assertTrue(len(self.sdp.moment_lowerbounds) == 4,
                        "Lower bounds are being reset when they should not.")
        self.assertTrue(len(self.sdp.moment_upperbounds) == 1,
                        "Upper bounds are being reset when they should not.")

    def test_reset_objective(self):
        self.prepare_objects(self.sdp)
        self.sdp.reset("objective")
        self.assertEqual(self.sdp.objective, {self.sdp.One: 0.},
                         "Resetting the objective function fails.")
        self.assertTrue(len(self.sdp.known_moments) == 3,
                        "Resetting the objective resets the known moments.")
        self.assertTrue(len(self.sdp.moment_lowerbounds) == 4,
                        "Resetting the objective resets the lower bounds.")
        self.assertTrue(len(self.sdp.moment_upperbounds) == 1,
                        "Resetting the objective resets the upper bounds.")

    def test_reset_values(self):
        self.prepare_objects(self.sdp)
        self.sdp.reset("values")
        self.assertEqual(self.sdp.semiknown_moments, {},
                         "Resetting the known values fails to empty " +
                         "semiknown_moments.")
        self.assertEqual(self.sdp.known_moments, {self.sdp.One: 1.},
                         "Resetting the known values fails to empty " +
                         "known_moments.")
        self.assertTrue(len(self.sdp.moment_lowerbounds) == 4,
                        "Resetting the objective resets the lower bounds.")
        self.assertTrue(len(self.sdp.moment_upperbounds) == 1,
                        "Resetting the objective resets the upper bounds.")
        self.assertTrue(len(self.sdp.objective) == 2,
                        "Resetting the bounds resets the objective function.")


class TestResetLP(unittest.TestCase):
    lp = InflationLP(trivial_c)
    lp._generate_lp()

    def setUp(self) -> None:
        var1 = "P[v_0=0 v_1=0]"
        var2 = "pv(0|1)"
        self.lp.set_objective({var1: 1}, "max")
        self.lp.set_bounds({var1: 0.9}, "up")
        self.lp.set_bounds({var1: 0.1}, "lo")
        self.lp.set_values({var2: 0.5})

    def test_reset_all(self):
        self.lp.reset("all")
        self.assertEqual(self.lp.moment_lowerbounds, {},
                         "Resetting lower bounds failed.")
        self.assertEqual(self.lp.moment_upperbounds, {},
                         "Resetting upper bounds failed.")
        self.assertEqual(self.lp.objective, {},
                         "Resetting objective failed.")
        self.assertEqual(self.lp.semiknown_moments, {},
                         "Resetting known values failed to empty "
                         "semiknown_moments.")
        self.assertEqual(self.lp.known_moments, {self.lp.One: 1.},
                         "Resetting known values failed to empty "
                         "known_moments.")


class TestSDPOutput(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore", category=UserWarning)

    def GHZ(self, v):
        dist = np.zeros((2, 2, 2, 1, 1, 1))
        for a in [0, 1]:
            for b in [0, 1]:
                for c in [0, 1]:
                    if (a == b) and (b == c):
                        dist[a, b, c, 0, 0, 0] = v/2 + (1-v)/8
                    else:
                        dist[a, b, c, 0, 0, 0] = (1-v)/8
        return dist

    cutInflation = InflationProblem({"lambda": ["a", "b"],
                                     "mu": ["b", "c"],
                                     "sigma": ["a", "c"]},
                                    outcomes_per_party=[2, 2, 2],
                                    settings_per_party=[1, 1, 1],
                                    inflation_level_per_source=[2, 1, 1])

    cutInflation_c = InflationProblem({"lambda": ["a", "b"],
                                       "mu": ["b", "c"],
                                       "sigma": ["a", "c"]},
                                      outcomes_per_party=[2, 2, 2],
                                      settings_per_party=[1, 1, 1],
                                      inflation_level_per_source=[2, 1, 1],
                                      classical_sources="all")

    instrumental = InflationProblem({"U_AB": ["A", "B"],
                                     "A": ["B"]},
                                    outcomes_per_party=(2, 2),
                                    settings_per_party=(3, 1),
                                    inflation_level_per_source=(1,),
                                    order=("A", "B"))
    instrumental_c = InflationProblem({"U_AB": ["A", "B"],
                                       "A": ["B"]},
                                      outcomes_per_party=(2, 2),
                                      settings_per_party=(3, 1),
                                      inflation_level_per_source=(1,),
                                      order=("A", "B"),
                                      classical_sources='all')

    incompatible_dist = np.array([[[[0.5], [0.5], [0.0]],
                                   [[0.0], [0.0], [0.5]]],
                                  [[[0.0], [0.5], [0.0]],
                                   [[0.5], [0.0], [0.5]]]], dtype=float)

    def test_bounds(self):
        ub = 0.8
        lb = 0.2
        very_trivial = InflationProblem({"a": ["b"]}, outcomes_per_party=[2])
        sdp          = InflationSDP(very_trivial)
        sdp.generate_relaxation("npa1")
        operator = np.asarray(very_trivial.measurements_symbolic).flatten()[0]
        sdp.set_objective(operator, "max")
        sdp.set_bounds({operator: ub}, "up")
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, ub),
                        "Setting upper bounds to monomials is failing. The " +
                        f"result obtained for [max x s.t. x <= {ub}] is " +
                        f"{sdp.objective_value}.")
        sdp.set_objective(operator, "min")
        sdp.set_bounds({operator: lb}, "lo")
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, lb),
                        "Setting upper bounds to monomials is failing. The " +
                        f"result obtained for [min x s.t. x >= {lb}] is " +
                        f"{sdp.objective_value}.")

    def test_CHSH(self):
        sdp = InflationSDP(bellScenario)
        sdp.generate_relaxation("npa1")
        self.assertEqual(sdp.n_columns, 5,
                         "The number of generating columns is not correct.")
        self.assertEqual(sdp.n_knowable, 8 + 1,  # only '1' is included here.
                         "The count of knowable moments is wrong.")
        self.assertEqual(sdp.n_unknowable, 2,
                         "The count of unknowable moments is wrong.")
        meas = bellScenario.measurements_symbolic
        A0 = 2*meas[0][0][0][0] - 1
        A1 = 2*meas[0][0][1][0] - 1
        B0 = 2*meas[1][0][0][0] - 1
        B1 = 2*meas[1][0][1][0] - 1

        sdp.set_objective(A0*(B0+B1)+A1*(B0-B1), "max")
        self.assertEqual(len(sdp.objective), 7,
                         "The parsing of the objective function is failing")
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, 2*np.sqrt(2)),
                        "The SDP is not recovering max(CHSH) = 2*sqrt(2)")
        bias = 3/4
        biased_chsh = 2.62132    # Value obtained by other means (ncpol2sdpa)
        sdp.set_values({meas[0][0][0][0]: bias,    # Variable for p(a=0|x=0)
                        "A_1_1_0": bias,           # Variable for p(a=0|x=1)
                        meas[1][0][0][0]: bias,    # Variable for p(b=0|y=0)
                        "B_1_1_0": bias            # Variable for p(b=0|y=1)
                        })
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, biased_chsh),
                        f"The SDP is not recovering max(CHSH) = {biased_chsh} "
                        + "when the single-party marginals are biased towards "
                        + str(bias))
        bias = 1/4
        biased_chsh = 2.55890
        sdp.set_values({meas[0][0][0][0]: bias,    # Variable for p(a=0|x=0)
                        "A_1_1_0": bias,           # Variable for p(a=0|x=1)
                        meas[1][0][0][0]: bias,    # Variable for p(b=0|y=0)
                        "B_1_1_0": bias            # Variable for p(b=0|y=1)
                        })
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, biased_chsh),
                        "The SDP is not re-setting the objective correctly " +
                        "after re-setting known values.")

    def test_equalities(self):
        prob = InflationProblem(dag={"U_AB": ["A", "B"],
                                     "U_AC": ["A", "C"],
                                     "U_AD": ["A", "D"],
                                     "U_CD": ["C", "D"],
                                     "C": ["D"],
                                     "A": ["B", "C", "D"]},
                                outcomes_per_party=(2, 2, 2, 2),
                                settings_per_party=(1, 1, 1, 1),
                                inflation_level_per_source=(1, 1, 1, 1),
                                order=("A", "B", "C", "D"))
        sdp = InflationSDP(prob)
        sdp.generate_relaxation("npa2")
        equalities = sdp.minimal_equalities

        self.assertEqual(len(equalities), 738,
                         "Failing to obtain the correct number of implicit " +
                         "equalities in a non-network scenario.")

        # TODO: When we add support for user-specifiable equalities, modify
        # this test to only check implicit equalities.
        self.assertTrue(all(set(equality.values()) == {-1, 1}
                            for equality in equalities),
                        "Some implicit equalities lack a nontrivial left-hand"
                        + "or right-hand side.")

        self.assertTrue(all(sdp.Zero.name not in equality.keys()
                            for equality in equalities),
                        "Some implicit equalities are wrongly assigning " +
                        "coefficients to the zero monomial.")

    def test_GHZ_commuting(self):
        sdp = InflationSDP(self.cutInflation_c)
        sdp.generate_relaxation("local1")
        self.assertEqual(sdp.n_columns, 18,
                         "The number of generating columns is not correct.")
        self.assertEqual(sdp.n_knowable, 8 + 1,  # only '1' is included here.
                         "The count of knowable moments is wrong.")
        self.assertEqual(sdp.n_unknowable, 11,
                         "The count of unknowable moments is wrong.")

        sdp.set_distribution(self.GHZ(0.5 + 1e-2))
        sdp.solve()
        self.assertEqual(sdp.status, "dual_infeas_cer",
                         "The commuting SDP is not identifying incompatible " +
                         "distributions.")
        lp_fanout = InflationLP(self.cutInflation_c)
        lp_fanout.set_distribution(self.GHZ(0.5 + 1e-2))
        lp_fanout.solve()
        self.assertEqual(lp_fanout.success, False,
                         "The fanout LP is not identifying incompatible " +
                         "distributions.")
        lp_nonfanout = InflationLP(self.cutInflation)
        lp_nonfanout.set_distribution(self.GHZ(0.5 + 1e-2))
        lp_nonfanout.solve()
        self.assertEqual(lp_nonfanout.success, False,
                         "The nonfanout LP is not identifying incompatible " +
                         "distributions.")
        sdp.solve(feas_as_optim=True)
        self.assertTrue(sdp.primal_objective <= 0,
                        "The commuting SDP with feasibility as optimization " +
                        "is not identifying incompatible distributions.")
        sdp.set_distribution(self.GHZ(0.5 - 1e-2))
        sdp.solve()
        self.assertEqual(sdp.status, "optimal",
                         "The commuting SDP is not recognizing compatible " +
                         "distributions.")
        lp_fanout.set_distribution(self.GHZ(0.5 - 1e-2))
        lp_fanout.solve()
        self.assertEqual(lp_fanout.success, True,
                         "The fanout LP is not recognizing compatible " +
                         "distributions.")
        lp_nonfanout.set_distribution(self.GHZ(0.5 - 1e-2))
        lp_nonfanout.solve()
        self.assertEqual(lp_nonfanout.success, True,
                         "The nonfanout LP is not identifying incompatible " +
                         "distributions.")
        sdp.solve(feas_as_optim=True)
        self.assertTrue(sdp.primal_objective >= 0,
                        "The commuting SDP with feasibility as optimization " +
                        "is not recognizing compatible distributions.")

    def test_GHZ_NC(self):
        sdp = InflationSDP(self.cutInflation)
        sdp.generate_relaxation("local1")
        self.assertEqual(sdp.n_columns, 18,
                         "The number of generating columns is not correct.")
        self.assertEqual(sdp.n_knowable, 8 + 1,  # only '1' is included here.
                         "The count of knowable moments is wrong.")
        self.assertEqual(sdp.n_unknowable, 13,
                         "The count of unknowable moments is wrong.")

        sdp.set_distribution(self.GHZ(0.5 + 1e-2))
        self.assertTrue(np.isclose(sdp.known_moments[sdp.moments[8]],
                                   (0.5+1e-2) / 2 + (0.5-1e-2) / 8),
                        "Setting the distribution is failing.")
        sdp.solve()
        self.assertTrue(sdp.status in ["dual_infeas_cer", "unknown"],
                        "The non-commuting SDP is not identifying " +
                        "incompatible distributions.")
        sdp.solve(feas_as_optim=True)
        self.assertTrue(sdp.primal_objective <= 0,
                        "The NC SDP with feasibility as optimization is not " +
                        "identifying incompatible distributions.")
        sdp.set_distribution(self.GHZ(0.5 - 1e-2))
        self.assertTrue(np.isclose(sdp.known_moments[sdp.moments[8]],
                                   (0.5-1e-2) / 2 + (0.5+1e-2) / 8),
                        "Re-setting the distribution is failing.")
        sdp.solve()
        self.assertEqual(sdp.status, "optimal",
                         "The non-commuting SDP is not recognizing " +
                         "compatible distributions.")
        sdp.solve(feas_as_optim=True)
        self.assertTrue(sdp.primal_objective >= 0,
                        "The NC SDP with feasibility as optimization is not " +
                        "recognizing compatible distributions.")

    def test_instrumental(self):
        sdp = InflationSDP(self.instrumental)
        sdp.generate_relaxation("local1")
        sdp.set_distribution(self.incompatible_dist)
        sdp.solve(feas_as_optim=False)
        self.assertEqual(sdp.status, "dual_infeas_cer",
                         "Failing to detect the infeasibility of the " +
                         "distribution that maximally violates Bonet's " +
                         "inequalty.")
        unnormalized_dist = np.ones((2, 2, 3, 1), dtype=float)
        sdp.set_distribution(unnormalized_dist)
        sdp.solve(feas_as_optim=False)
        self.assertEqual(sdp.status, "dual_infeas_cer",
                         "Failing to detect the infeasibility of an " +
                         "distribution that violates normalization.")
        compat_dist = unnormalized_dist / 4
        sdp.set_distribution(compat_dist)
        sdp.solve(feas_as_optim=False)
        self.assertEqual(sdp.status, "optimal",
                         "A feasible distribution for the instrumental " +
                         "scenario is not being recognized as such.")

    def test_lpi(self):
        sdp = InflationSDP(trivial)
        [[[[A10,_],[A11,_]],[[A20,_],[A21,_]]]] = trivial.measurements_symbolic
        sdp.generate_relaxation([1,
                                 A10, A11, A20, A21,
                                 A10*A11, A10*A21, A11*A20, A20*A21])
        sdp.set_distribution(np.array([[0.14873, 0.85168], [None, None]]))
        sdp.set_objective(A11*A10*A20*A21)
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, 0.0918999),
                        "Optimization of a simple SDP without LPI-like " +
                        "constraints is not obtaining the correct known value."
                        )
        sdp.set_distribution(np.array([[0.14873, 0.85168], [None, None]]),
                             use_lpi_constraints=True
                             )
        sdp.solve()
        self.assertTrue(np.isclose(sdp.objective_value, 0.0640776),
                        "Optimization of a simple SDP with LPI-like " +
                        "constraints is not obtaining the correct known value."
                        )

    def test_lpi_bounds(self):
        sdp  = InflationSDP(trivial)
        cols = [np.array([]),
                np.array([[1, 2, 0, 0],
                          [1, 2, 1, 0]]),
                np.array([[1, 1, 0, 0],
                          [1, 2, 0, 0],
                          [1, 2, 1, 0]])]
        sdp.generate_relaxation(cols)
        sdp.set_distribution(np.ones((2, 2)) / 2,
                             use_lpi_constraints=True)

        self.assertGreaterEqual(len(sdp.semiknown_moments), 1,
                                "Failing to identify semiknowns.")

        self.assertTrue(all(abs(val[0]) <= 1.
                            for val in sdp.semiknown_moments.values()),
                        "Semiknown moments need to be of the form " +
                        "mon_index1 = (number<=1) * mon_index2, this is " +
                        "failing.")

    def test_new_indices(self):
        sdp  = InflationSDP(trivial)
        cols = [np.array([]),
                np.array([[1, 1, 0, 0],
                          [1, 2, 0, 0],
                          [1, 2, 1, 0]])]
        sdp.generate_relaxation(cols)
        sdp.set_distribution(np.ones((2, 2)) / 2,
                             use_lpi_constraints=True)
        new_mon_indices = np.array([semi[1][1].idx
                                    for semi in sdp.semiknown_moments.items()])
        self.assertTrue(np.all(new_mon_indices > len(sdp.moments)),
                        "The new unknown monomials that appear when applying" +
                        " LPI constraints are not assigned correct indices.")

    def test_supports(self):
        sdp = InflationSDP(bellScenario, supports_problem=True)
        sdp.generate_relaxation("local1")
        pr_support = np.zeros((2, 2, 2, 2))
        for a, b, x, y in np.ndindex(*pr_support.shape):
            if x*y == (a + b) % 2:
                pr_support[a, b, x, y] = np.random.randn()
        sdp.set_distribution(pr_support)
        sdp.solve(feas_as_optim=False)
        self.assertEqual(sdp.status, "dual_infeas_cer",
                         "Failing to detect the infeasibility of a support "
                         + "known to be incompatible.")
        compatible_support = np.ones((2, 2, 2, 2), dtype=float)
        sdp.set_distribution(compatible_support)
        sdp.solve(feas_as_optim=False)
        self.assertEqual(sdp.status, "optimal",
                         "A feasible support for the Bell scenario is not " +
                         "being recognized as such.")


class TestLPOutput(unittest.TestCase):
    def test_bounds(self):
        ub = 0.8
        lb = 0.2
        very_trivial = InflationProblem({"a": ["b"]}, outcomes_per_party=[2])
        lp = InflationLP(very_trivial)
        operator = np.asarray(lp.monomials).flatten()[1]
        with self.subTest(msg="Setting upper bound"):
            lp.set_objective({operator: 1}, "max")
            lp.set_bounds({operator: ub}, "up")
            lp.solve()
            self.assertAlmostEqual(
                lp.objective_value, ub,
                msg=f"Setting upper bounds to monomials failed. The result "
                    f"obtained for {{max x s.t. x <= {ub}}} is "
                    f"{lp.objective_value}.")
        with self.subTest(msg="Setting lower bound"):
            lp.set_objective({operator: 1}, "min")
            lp.set_bounds({operator: lb}, "lo")
            lp.solve()
            self.assertAlmostEqual(
                lp.objective_value, lb,
                msg=f"Setting lower bounds to monomials failed. The result "
                    f"obtained for {{min x s.t. x >= {lb}}} is "
                    f"{lp.objective_value}.")

    def test_instrumental(self):
        lp = InflationLP(TestSDPOutput.instrumental_c)
        with self.subTest(msg="Infeasible Bonet's inequality"):
            lp.set_distribution(TestSDPOutput.incompatible_dist)
            lp.solve(feas_as_optim=False)
            self.assertTrue(lp.status in ["prim_infeas_cer", "dual_infeas_cer",
                                          "unknown"],
                            "Failed to detect the infeasibility of the "
                            "distribution that maximally violates Bonet's "
                            "inequality.")
        with self.subTest(msg="Infeasible normalization"):
            unnormalized_dist = np.ones((2, 2, 3, 1), dtype=float)
            lp.set_distribution(unnormalized_dist)
            lp.solve(feas_as_optim=False)
            self.assertTrue(lp.status in ["prim_infeas_cer", "dual_infeas_cer",
                                          "unknown"],
                            "Failed to detect the infeasibility of a "
                            "distribution that violates normalization.")
        with self.subTest(msg="Feasible instrumental"):
            compatible_dist = unnormalized_dist / 4
            lp.set_distribution(compatible_dist)
            lp.solve(feas_as_optim=False)
            self.assertEqual(lp.status, "optimal",
                             "A feasible distribution for the instrumental "
                             "scenario is not being recognized as such.")

    def test_supports(self):
        lp = InflationLP(bellScenario_c, supports_problem=True)
        with self.subTest(msg="Incompatible support"):
            pr_support = np.zeros((2, 2, 2, 2))
            for a, b, x, y in np.ndindex(*pr_support.shape):
                if x*y == (a + b) % 2:
                    pr_support[a, b, x, y] = np.random.randn()
            lp.set_distribution(pr_support)
            lp.solve(feas_as_optim=False)
            self.assertIn(lp.status, ["prim_infeas_cer", "dual_infeas_cer"],
                          "Failed to detect the infeasibility of a support "
                          "known to be incompatible.")
        with self.subTest(msg="Compatible support"):
            compatible_support = np.ones((2, 2, 2, 2), dtype=float)
            lp.set_distribution(compatible_support)
            lp.solve(feas_as_optim=False)
            self.assertEqual(lp.status, "optimal",
                             "A feasible support for the Bell scenario was "
                             "not recognized as such.")


class TestSymmetries(unittest.TestCase):
    def test_apply_symmetries(self):
        from inflation.sdp.quantum_tools import \
                                                apply_inflation_symmetries
        G = np.array([[0,  1,  2,  3,  4,  5],
                      [6,  7,  8,  9, 10, 11],
                      [12, 13, 14, 15, 16, 17],
                      [18, 19, 20, 21, 22, 23],
                      [24, 25, 26, 27, 28, 29],
                      [30, 31, 32, 33, 34, 35]])
        # The symmetries are that equal variables are (1,2), (3, 4, 5), (8, 9)
        sym_mm, orbits, repr_values = \
            apply_inflation_symmetries(G, np.array([[0, 1, 2, 3, 5, 4],
                                                    [0, 2, 3, 1, 4, 5]]))
        sym_mm_good = np.array([[0,  1,  2,  1,  3,  3],
                                [4,  5,  6,  7,  8,  8],
                                [9, 10, 11, 12, 13, 13],
                                [4,  6,  7,  5,  8,  8],
                                [14, 15, 16, 15, 17, 18],
                                [14, 15, 16, 15, 18, 17]])
        self.assertTrue(np.allclose(sym_mm, sym_mm_good),
                        "Symmetrized moment matrix is not correct.")
        orbits_good = {0: 0, 1: 1, 2: 2, 3: 1, 4: 3, 5: 3, 6: 4, 7: 5, 8: 6,
                       9: 7, 10: 8, 11: 8, 12: 9, 13: 10, 14: 11, 15: 12,
                       16: 13, 17: 13, 18: 4, 19: 6, 20: 7, 21: 5, 22: 8,
                       23: 8, 24: 14, 25: 15, 26: 16, 27: 15, 28: 17, 29: 18,
                       30: 14, 31: 15, 32: 16, 33: 15, 34: 18, 35: 17}
        self.assertTrue(orbits == orbits_good,
                        "Orbits dictionary is not correct.")
        repr_values_good = np.array([0,  1,  2,  4,  6,  7,  8,  9, 10,
                                     12, 13, 14, 15, 16, 24, 25, 26, 28, 29])
        self.assertTrue(np.allclose(repr_values, repr_values_good),
                        "Representatives mapping is not correct.")

    def test_commutations_after_symmetrization(self):
        scenario = InflationSDP(trivial_c)
        col_structure = [np.array([]),
                         np.array([[1, 2, 0, 0], [1, 2, 1, 0]]),
                         np.array([[1, 1, 0, 0], [1, 2, 0, 0]]),
                         np.array([[1, 1, 1, 0], [1, 2, 0, 0]]),
                         np.array([[1, 1, 0, 0], [1, 2, 1, 0]]),
                         np.array([[1, 1, 1, 0], [1, 2, 1, 0]]),
                         np.array([[1, 1, 0, 0], [1, 1, 1, 0]])]

        scenario.generate_relaxation(col_structure)
        self.assertTrue(np.array_equal(scenario.columns_symmetries,
                                       [[0, 6, 2, 4, 3, 5, 1]]),
                        "The commutation relations of different copies are " +
                        "not applied properly after inflation symmetries.")

    def test_detected_symmetries(self):
        bilocalSDP.build_columns('local1')
        syms = bilocalSDP._discover_columns_symmetries()
        # Make it a set so the order doesn't matter
        syms = set(tuple(s) for s in syms)
        # I simply copied the output at a time when we understand the code
        # to be working; this test simply detects if the code changes
        syms_good = {(0, 1, 2, 4, 3, 6, 5, 8, 7, 10, 9, 12, 11, 14,
                      13, 16, 15, 18, 17, 20, 19, 24, 23, 22, 21, 28,
                      27, 26, 25, 32, 31, 30, 29, 36, 35, 34, 33, 40,
                      39, 38, 37, 44, 43, 42, 41),
                     (0, 2, 1, 5, 6, 3, 4, 7, 8, 15, 16, 13, 14, 11,
                      12, 9, 10, 19, 20, 17, 18, 25, 26, 27, 28, 21,
                      22, 23, 24, 41, 42, 43, 44, 37, 38, 39, 40, 33,
                      34, 35, 36, 29, 30, 31, 32),
                     (0, 2, 1, 6, 5, 4, 3, 8, 7, 16, 15, 14, 13, 12,
                      11, 10, 9, 20, 19, 18, 17, 28, 27, 26, 25, 24,
                      23, 22, 21, 44, 43, 42, 41, 40, 39, 38, 37, 36,
                      35, 34, 33, 32, 31, 30, 29)}
        self.assertEqual(syms, syms_good, "The symmetries are not being " +
                                          "detected correctly.")

    def test_symmetrized_PRbox(self):

        from inflation.symmetry_utils import discover_distribution_symmetries

        def PR_box(v):
            prob = np.zeros((2,2,2,2), dtype=float)
            for a,b,x,y in np.ndindex(*prob.shape):
                if np.bitwise_xor(a,b) == np.bitwise_and(x,y):
                    prob[a,b,x,y] = 0.5

            return v * prob + (1-v) * np.ones_like(prob) / 4

        symmetries = discover_distribution_symmetries(PR_box(1), bellScenario)
        bellScenario.add_symmetries(symmetries)
        bellScenario_c.add_symmetries(symmetries)

        BellLP = InflationLP(bellScenario_c)
        BellLP.set_distribution(PR_box(1/2+1e-4))
        BellLP.solve()
        self.assertEqual(BellLP.success, False,
                         "The symmetrized LP is not identifying incompatible" +
                         " distributions.")
        BellLP.set_distribution(PR_box(1/2-1e-4))
        BellLP.solve()
        self.assertEqual(BellLP.status, "optimal",
                         "The symmetrized LP is not identifying compatible" +
                         " distributions.")

        BellSDP = InflationSDP(bellScenario)
        BellSDP.generate_relaxation("npa2")
        BellSDP.set_distribution(PR_box(1/np.sqrt(2)+5e-3))
        BellSDP.solve()
        self.assertEqual(BellSDP.status, "dual_infeas_cer",
                         "The symmetrized SDP is not identifying incompatible" +
                         " distributions.")
        BellSDP.set_distribution(PR_box(1/np.sqrt(2)-1e-4))
        BellSDP.solve()
        self.assertEqual(BellSDP.status, "optimal",
                         "The symmetrized SDP is not identifying compatible" +
                         " distributions.")

        bellScenario.reset_symmetries()
        bellScenario_c.reset_symmetries()


class TestConstraintGeneration(unittest.TestCase):
    def test_norm_eqs_mon2index_mapping(self):
        sdp = InflationSDP(InflationProblem({'r': ['A']}, (3,), (3,), (3,)))
        sdp.generate_relaxation('npa2')
        for i, lexmon in enumerate(sdp.generating_monomials_1d):
            self.assertTrue(i == sdp.genmon_1d_to_index[tuple(lexmon)],
                            "Monomials in the generating list must be mapped "
                            + "to their index in the list under "
                            + "InflationSDP.genmon_hash_to_index for "
                            + "InflationSDP._discover_normalization_eqns() "
                            + "to work correctly.")

    def test_norm_eqs_expansion(self):
        from inflation.sdp.quantum_tools import \
                                                expand_moment_normalisation

        out = expand_moment_normalisation(np.array([[1, 1, 0, 1]]),
                                          [3],
                                          {i: False for i in range(3)})
        out_good = [(np.array([]).reshape((0, 4)),
                    [np.array([[1, 1, 0, 1]]), np.array([[1, 1, 0, 0]])])]
        for eq1, eq2 in zip(out, out_good):
            self.assertTrue(np.allclose(eq1[0], eq2[0])
                            and len(eq1[1]) == len(eq2[1]),
                            "Normalisation constraint is not" +
                            "being properly generated. ")
            for el1, el2 in zip(eq1[1], eq2[1]):
                self.assertTrue(np.allclose(el1, el2),
                                "Normalisation constraint is not" +
                                "being properly generated. ")

        # Note, currently if more than one operator has a last output,
        # several equalities will be generated, instead of a single one.
        out = expand_moment_normalisation(np.array([[1, 1, 0, 1],
                                                    [2, 3, 1, 1],
                                                    [2, 3, 1, 0]]),
                                          [3, 3, 2],  # we 'lie' about card.
                                          {i: False for i in range(3)})
        out_good = [(np.array([[2, 3, 1, 1], [2, 3, 1, 0]]),
                    [np.array([[1, 1, 0, 1], [2, 3, 1, 1], [2, 3, 1, 0]]),
                     np.array([[1, 1, 0, 0], [2, 3, 1, 1], [2, 3, 1, 0]])]),
                    (np.array([[1, 1, 0, 1], [2, 3, 1, 0]]),
                    [np.array([[1, 1, 0, 1], [2, 3, 1, 1], [2, 3, 1, 0]]),
                     np.array([[1, 1, 0, 1], [2, 3, 1, 0], [2, 3, 1, 0]])])]
        for eq1, eq2 in zip(out, out_good):
            self.assertTrue(np.allclose(eq1[0], eq2[0])
                            and len(eq1[1]) == len(eq2[1]),
                            "Normalisation constraint is not" +
                            "being properly generated. ")
            for el1, el2 in zip(eq1[1], eq2[1]):
                self.assertTrue(np.allclose(el1, el2),
                                "Normalisation constraint is not" +
                                "being properly generated.")

    def test_normeqs_colineq2momentineq(self):
        from inflation.sdp.quantum_tools import \
                                            construct_normalization_eqs
        G = np.array([[1,  2,  3,  4,  5],
                      [2,  1,  6,  7,  8],
                      [3,  6,  1,  9, 10],
                      [4,  7,  9,  1, 11],
                      [5,  8, 10, 11,  1]])
        # The following inequalities have random integers, they have no
        # interpretation
        column_inequalities = [(0, [1, 2, 3, 4]), (3, [2, 4])]
        out = construct_normalization_eqs(column_inequalities, G)
        out_good = [(1, [2, 3, 4, 5]),
                    (2, [1, 6, 7, 8]),
                    (3, [1, 6, 9, 10]),
                    (4, [1, 7, 9, 11]),
                    (5, [1, 8, 10, 11]),
                    (4, [3, 5]),
                    (7, [6, 8]),
                    (9, [1, 10]),
                    (1, [9, 11])]
        self.assertEqual(out,
                         out_good,
                         "Column equalities are not being " +
                         "properly lifted to moment inequalities.")


class TestPipelineLP(unittest.TestCase):
    def _monomial_generation(self, **args):
        with self.subTest(msg="Testing monomial generation"):
            raw_n_columns = args["scenario"].raw_n_columns
            truth = args["truth_columns"]
            self.assertEqual(raw_n_columns, truth,
                             f"There are {raw_n_columns} columns but {truth} "
                             f"were expected.")

    def _equalities(self, **args):
        with self.subTest(msg="Testing equalities"):
            equalities = args["scenario"].moment_equalities
            truth = args["truth_eq"]
            self.assertEqual(len(equalities), truth,
                             f"There are {len(equalities)} equalities but "
                             f"{truth} were expected.")
            self.assertTrue(all(set(equality.values()) == {-1, 1}
                                for equality in equalities),
                            "Some implicit equalities lack a nontrivial "
                            "left-hand or right-hand side.")

    def _test_a_visibility(self, **args):
        lp = args["scenario"]
        dist_func = args["dist_func"]
        dist_name = args["dist_name"]
        crit_cutoff = args["crit_cutoff"]
        with self.subTest(msg="Testing GHZ, incompatible distribution"):
            lp.set_distribution(dist_func(crit_cutoff + 1e-2))
            lp.solve()
            self.assertIn(lp.status, ["prim_infeas_cer", "dual_infeas_cer",
                                      "unknown"],
                          "The LP did not identify the incompatible "
                          "distribution.")
        with self.subTest(msg=f"Testing {dist_name}, incompatible distribution, "
                              "feasibility as optimization"):
            lp.solve(feas_as_optim=True)
            self.assertTrue(lp.primal_objective <= 0,
                            "The LP with feasibility as optimization did not "
                            "identify the incompatible distribution.")
        with self.subTest(msg=f"Testing {dist_name}, compatible distribution"):
            lp.set_distribution(dist_func(crit_cutoff - 1e-2))
            lp.solve()
            self.assertEqual(lp.status, "optimal",
                             "The LP did not recognize the compatible "
                             "distribution.")
        with self.subTest(msg=f"Testing {dist_name}, compatible distribution, "
                              "feasibility as optimization"):
            lp.solve(feas_as_optim=True)
            self.assertTrue(lp.primal_objective >= -1e-6,
                            "The LP with feasibility as optimization did not "
                            "recognize the compatible distribution.")

    def _run(self, **args):
        self._monomial_generation(**args)
        self._equalities(**args)
        self._test_a_visibility(**args)


class TestInstrumental(TestPipelineLP):
    instrumental = InflationProblem({"U_AB": ["A", "B"],
                                     "A": ["B"]},
                                    outcomes_per_party=(2, 2),
                                    settings_per_party=(2, 1),
                                    inflation_level_per_source=(1,),
                                    order=("A", "B"))
    instrumental_c = InflationProblem({"U_AB": ["A", "B"],
                                       "A": ["B"]},
                                      outcomes_per_party=(2, 2),
                                      settings_per_party=(2, 1),
                                      inflation_level_per_source=(1,),
                                      order=("A", "B"),
                                      classical_sources='all')

    def p_Pearl_Violating(self, v):
        dist = np.full((2, 2, 2, 1), (1 - v) / 4)
        dist[0, 0, 0, 0] = dist[0, 1, 1, 0] = v + (1 - v) / 4
        return dist

    def test_instrumental_fanout(self):
        inst = InflationLP(self.instrumental_c)
        args = {"scenario": inst,
                "truth_columns": 36,
                "truth_eq": 20,
                "dist_func": self.p_Pearl_Violating,
                "dist_name": "Instrumental noisy e-sep violating",
                "crit_cutoff": 1/3}
        self._run(**args)

    def test_instrumental_nonfanout(self):
        inst = InflationLP(self.instrumental)
        args = {"scenario": inst,
                "truth_columns": 15,
                "truth_eq": 6,
                "dist_func": self.p_Pearl_Violating,
                "dist_name": "Instrumental noisy e-sep violating",
                "crit_cutoff": 1/3}
        self._run(**args)


class TestBell(TestPipelineLP):

    def _CHSH(self, **args):
        lp = args["scenario"]
        truth = args["truth_obj"]
        lp.set_objective({1: 2.0,
                          "pA(0|0)": -4.0,
                          "pB(0|0)": -4.0,
                          "pAB(00|11)": -4.0,
                          "pAB(00|00)": 4.0,
                          "pAB(00|01)": 4.0,
                          "pAB(00|10)": 4.0}, "max")
        with self.subTest(msg="Testing max CHSH"):
            lp.solve()
            self.assertAlmostEqual(lp.objective_value, truth,
                                   msg=f"The LP is not recovering max(CHSH) = "
                                       f"{truth}.")

    def p_Signalling_to_Bob(self, v):
        dist = np.full((2, 2, 2, 2), (1 - v) / 4)
        dist[0, 0, 0, 0] = dist[0, 1, 1, 0] = v + (1 - v) / 4
        return dist

    def test_bell_fanout(self):
        bell = InflationLP(bellScenario_c)
        args = {"scenario": bell,
                "truth_columns": 16,
                "truth_obj": 2,
                "truth_eq": 0,
                "dist_func": self.p_Signalling_to_Bob,
                "dist_name": "Noisy Signalling to Bob",
                "crit_cutoff": 0.2}
        self._run(**args)

    def test_bell_nonfanout(self):
        bell = InflationLP(bellScenario)
        args = {"scenario": bell,
                "truth_columns": 9,
                "truth_obj": 2,
                "truth_eq": 0,
                "dist_func": self.p_Signalling_to_Bob,
                "dist_name": "Noisy Signalling to Bob",
                "crit_cutoff": 0.2}
        self._run(**args)


class TestTriangle(TestPipelineLP):
    triangle = InflationProblem({"lambda": ["a", "b"],
                                 "mu": ["b", "c"],
                                 "sigma": ["a", "c"]},
                                outcomes_per_party=[2, 2, 2],
                                settings_per_party=[1, 1, 1],
                                inflation_level_per_source=[1, 1, 1],
                                order=['a', 'b', 'c'])
    
    triangle_c = InflationProblem({"lambda": ["a", "b"],
                                   "mu": ["b", "c"],
                                   "sigma": ["a", "c"]},
                                  outcomes_per_party=[2, 2, 2],
                                  settings_per_party=[1, 1, 1],
                                  inflation_level_per_source=[1, 1, 1],
                                  order=['a', 'b', 'c'],
                                  classical_sources="all")

    def GHZ(self, v):
        dist = np.zeros((2, 2, 2, 1, 1, 1))
        dist[0, 0, 0] = dist[1, 1, 1] = 1 / 2
        return v * dist + (1 - v) / 8

    def test_triangle_fanout(self):
        triangle = InflationLP(self.triangle_c)
        args = {"scenario": triangle,
                "truth_columns": 8,
                "truth_eq": 0,
                "dist_func": self.GHZ,
                "dist_name": "Noisy GHZ",
                "crit_cutoff": 1}
        self._run(**args)

    def test_triangle_nonfanout(self):
        triangle = InflationLP(self.triangle)
        args = {"scenario": triangle,
                "truth_columns": 8,
                "truth_eq": 0,
                "dist_func": self.GHZ,
                "dist_name": "Noisy GHZ",
                "crit_cutoff": 1}
        self._run(**args)


class TestEvans(TestPipelineLP):
    evans = InflationProblem({"U_AB": ["A", "B"],
                              "U_BC": ["B", "C"],
                              "B": ["A", "C"]},
                             outcomes_per_party=(2, 2, 2),
                             settings_per_party=(1, 1, 1),
                             inflation_level_per_source=(1, 1),
                             order=("A", "B", "C"))
    evans_c = InflationProblem({"U_AB": ["A", "B"],
                                "U_BC": ["B", "C"],
                                "B": ["A", "C"]},
                               outcomes_per_party=(2, 2, 2),
                               settings_per_party=(1, 1, 1),
                               inflation_level_per_source=(1, 1),
                               order=("A", "B", "C"),
                               classical_sources='all')

    def p_Evans_esep_violating(self, v):
        dist = np.zeros((2, 2, 2, 1, 1, 1))
        for x, y, z in product(range(2), repeat=3):
            dist[x, y, z] = (1 + v * (-1) ** (x + y + y * z)) / 8
        return dist

    def test_evans_fanout(self):
        evans = InflationLP(self.evans_c)
        args = {"scenario": evans,
                "truth_columns": 48,
                "truth_eq": 16,
                "dist_func": self.p_Evans_esep_violating,
                "dist_name": "Noisy Evans Incompatible",
                "crit_cutoff": 1}
        self._run(**args)

    def test_evans_nonfanout(self):
        evans = InflationLP(self.evans)
        args = {"scenario": evans,
                "truth_columns": 27,
                "truth_eq": 9,
                "dist_func": self.p_Evans_esep_violating,
                "dist_name": "Noisy Evans Incompatible",
                "crit_cutoff": 1}
        self._run(**args)
        

class TestFullNN(TestPipelineLP):
    scenario = InflationProblem({'lambda': ['A', 'B'],
                                     'NS': ['B', 'C']},
                                [2, 4, 2], [3, 1, 3], 
                                inflation_level_per_source=[1, 2],
                                classical_sources=['lambda'],)

    def _prob_EJM(self, v, theta=0):
        p_bell = np.expand_dims((0, 1, -1, 0), axis=1)/np.sqrt(2)
        rho_v = v * p_bell @ p_bell.conj().T + (1 - v) * np.eye(4)/4
        sigmax = np.array([[0, 1], [1, 0]])
        sigmay = np.array([[0, -1j], [1j, 0]])
        sigmaz = np.array([[1, 0], [0, -1]])
        A = [[np.expand_dims(v, axis=1) @ np.expand_dims(v, axis=1).conj().T 
                for v in reversed(np.linalg.eigh(op)[1].T)] 
                for op in [sigmax, sigmay, sigmaz]]
        C = [[np.expand_dims(v, axis=1) @ np.expand_dims(v, axis=1).conj().T 
                for v in reversed(np.linalg.eigh(op)[1].T)] 
                for op in [sigmax, sigmay, sigmaz]]
        r_plus = (1 + np.exp(1j*theta))/np.sqrt(2)
        r_minus = (1 - np.exp(1j*theta))/np.sqrt(2)
        e00 = np.expand_dims([1, 0, 0, 0], axis=1)
        e01 = np.expand_dims([0, 1, 0, 0], axis=1)
        e10 = np.expand_dims([0, 0, 1, 0], axis=1)
        e11 = np.expand_dims([0, 0, 0, 1], axis=1)
        psi1 = 1/2 * (np.exp(-1j*np.pi/4)*e00 - r_plus * e01 
                        - r_minus * e10 + np.exp(-3/4*np.pi*1j)*e11)
        psi2 = 1/2 * (np.exp(1j*np.pi/4)*e00 + r_minus * e01 
                        + r_plus * e10 + np.exp(3/4*np.pi*1j)*e11)
        psi3 = 1/2 * (np.exp(-1j*np.pi*3/4)*e00 + r_minus * e01 
                        + r_plus * e10 + np.exp(-1/4*np.pi*1j)*e11)
        psi4 = 1/2 * (np.exp(1j*np.pi*3/4)*e00 - r_plus * e01 
                        - r_minus * e10 + np.exp(np.pi*1j/4)*e11)
        B = [ [psi @ psi.conj().T for psi in [psi1, psi2, psi3, psi4] ]]
        
        p = np.zeros((2, 4, 2, 3, 1, 3))
        state = np.kron(rho_v, rho_v)
        for a, b, c, x, y, z in np.ndindex(p.shape):
            mmnt = np.kron(np.kron(A[x][a], B[y][b]), C[z][c])
            p[a, b, c, x, y, z] = np.real(np.trace(state @ mmnt))
        return p
    
    def test_fullnetworknonlocality_3partite_line(self):
        lp = InflationLP(self.scenario)
        
        best_theta = np.arccos(np.sqrt(5) / 3)
        v_for_best_theta = 2 / np.sqrt(5)
        
        args = {"scenario": lp,
                "truth_columns": 2048,
                "truth_eq": 0,
                "dist_func": lambda x: self._prob_EJM(x, theta=best_theta),
                "dist_name": "Noisy EJM",
                "crit_cutoff": v_for_best_theta}
        self._run(**args)